#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

static char buffer1[1000];
static char buffer2[1000];

FILE *infile;
FILE *outfile;

#ifdef __WATCOMC__
#define NEW_LINE "\n"
#else
#define NEW_LINE "\r\n"
#endif

void ReplaceString(char *str1, char *str2)
{
    char *instr = buffer1;
    char *outstr = buffer2;
    while (*instr)
    {
        if (!strncmp(instr, str1, strlen(str1)))
        {
            strcpy(outstr, str2);
            instr += strlen(str1);
            outstr += strlen(str2);
        }
        else
            *outstr++ = *instr++;
    }
    *outstr = 0;
    strcpy(buffer1, buffer2);
}

void ProcessAscii(void)
{
    int quote = 0;
    int comma = 0;
    char *instr = buffer1;
    char *outstr = buffer2;

    strcpy(outstr, "\tbyte ");
    outstr += 6;
    instr += 8;

    while (*instr)
    {
        if (*instr == '"')
        {
            if (quote)
            {
                quote = 0;
                *outstr++ = *instr++;
                continue;
            }
            else
            {
                instr++;
                continue;
            }
        }
        if (*instr == '\\')
        {
            int value = 0;
            if (quote)
            {
                *outstr++ = '"';
                quote = 0;
                comma = 1;
            }
            if (comma)
            {
                comma = 0;
                *outstr++ = ',';
                *outstr++ = ' ';
            }
            instr++;
            if (*instr >= '0' && *instr <= '7')
            {
                while (*instr >= '0' && *instr <= '7')
                {
                    value = (value << 3) + *instr - '0';
                    instr++;
                }
            }
            else
            {
                if (*instr == 't') value = 9;
                else if (*instr == 'n') value = 10;
                else if (*instr == 'r') value = 13;
                else value = *instr;
                instr++;
            }
            sprintf(outstr, "%d", value);
            outstr += strlen(outstr);
            comma = 1;
            continue;
        }
        if (!quote)
        {
            quote = 1;
            if (comma)
            {
                *outstr++ = ',';
                comma = 0;
            }
            *outstr++ = '"';
        }
        *outstr++ = *instr++;
    }
    *outstr = 0;
    strcpy(buffer1, buffer2);
}

char *FindChar(char *str, int val)
{
    while (*str)
    {
        if (*str == val) break;
        str++;
    }
    return str;
}

char *FindString(char *str1, char *str2)
{
    while (*str1)
    {
        if (!strncmp(str1, str2, strlen(str2))) break;
        str1++;
    }
    return str1;
}

char *SkipWhiteSpace(char *ptr)
{
    while (*ptr)
    {
        if (*ptr != ' ' && *ptr != '\t') break;
        ptr++;
    }
    return ptr;
}

char *FindWhiteSpace(char *ptr)
{
    while (*ptr)
    {
        if (*ptr == ' ' || *ptr == '\t') break;
        ptr++;
    }
    return ptr;
}

int CheckMova(void)
{
    char *ptr1, *ptr2;

    ptr1 = FindString(buffer1, "mova");
    if (!*ptr1) return 0;

    ptr2 = FindChar(ptr1, '#');
    if (!*ptr2) return 0;
    ptr1[3] = 0;
    ptr2[0] = 0;
    fprintf(outfile, "%s%s##%s%s", buffer1, ptr1+4, ptr2+1, NEW_LINE);
    return 1;
}

int CheckWaitcnt(void)
{
    char *ptr1, *ptr2;

    ptr1 = FindString(buffer1, "waitcnt");
    if (!*ptr1) return 0;
    ReplaceString("waitcnt", "addct1");
    fprintf(outfile, "%s%s", buffer1, NEW_LINE);
    fprintf(outfile, "\twaitct1%s", NEW_LINE);
    return 1;
}

int CheckCNT(void)
{
    char *ptr1, *ptr2;
    int len = strlen(buffer1);

    if (len < 7) return 0;
    if (strcmp(&buffer1[len-5], ", CNT")) return 0;
    buffer1[len-5] = 0;
    if (!strncmp(buffer1, "\tmov\t", 5))
    {
        fprintf(outfile, "\tgetct\t%s%s", &buffer1[5], NEW_LINE);
    }
    else
    {
        fprintf(outfile, "\tgetct\ttemp%s", NEW_LINE);
        fprintf(outfile, "%s, temp%s", buffer1, NEW_LINE);
    }
    return 1;
}

int CheckNR(void)
{
    char *ptr1, *ptr2;
    int len = strlen(buffer1);

    if (len < 5) return 0;
    if (strcmp(&buffer1[len-3], ",nr")) return 0;

    ptr1 = SkipWhiteSpace(buffer1);
    if (!*ptr1) return 0;
    if (!strncmp(ptr1, "IF_", 3))
    {
        ptr1 = FindWhiteSpace(ptr1);
        ptr1 = SkipWhiteSpace(ptr1);
        if (!*ptr1) return 0;
    }
    ptr1 = FindWhiteSpace(ptr1);
    ptr1 = SkipWhiteSpace(ptr1);
    if (!*ptr1) return 0;
    ptr2 = FindChar(ptr1, ',');
    if (!*ptr2) return 0;
    *ptr2++ = 0;
    buffer1[len-3] = 0;
    fprintf(outfile, "\tmov\ttemp, %s%s", ptr1, NEW_LINE);
    *ptr1 = 0;
    fprintf(outfile, "%stemp,%s%s", buffer1, ptr2, NEW_LINE);
    return 1;
}

int CheckSourceDest(void)
{
    int len;
    char *cptr, *fptr, *pptr, *optr;
    char *lptr = FindString(buffer1, "_L");
    char first[100];

    if (!*lptr) return 0;
    if (lptr == buffer1) return 0;

    fptr = SkipWhiteSpace(buffer1);
    if (!strncmp(fptr, "IF_", 3))
    {
        optr = FindWhiteSpace(fptr);
        optr = SkipWhiteSpace(optr);
    }
    else
        optr = fptr;
    if (!*optr) return 0;
    pptr = FindWhiteSpace(optr);
    if (!*pptr) return 0;

    if (!strncmp(optr, "jmp", 3)) return 0;
    if (!strncmp(optr, "djnz", 4)) return 0;
    if (!strncmp(optr, "call", 4)) return 0;
    if (!strncmp(optr, "calld", 5)) return 0;
    if (!strncmp(optr, "long", 4)) return 0;

    *pptr++ = 0;

    len = (int)optr - (int)buffer1;
    memcpy(first, buffer1, len);
    first[len] = 0;

    cptr = FindChar(pptr, ',');
    if (*cptr) *cptr++ = 0;

    if (!strcmp(optr, "mov"))
    {
        if ((int)lptr >= (int)cptr)
            fprintf(outfile, "%srdlong\t%s, ##%s%s", first, pptr, lptr, NEW_LINE);
        else
            fprintf(outfile, "%swrlong\t%s, ##%s%s", first, pptr, lptr, NEW_LINE);
    }
    else
    {
        fprintf(outfile, "%srdlong\ttemp, ##%s%s", first, lptr, NEW_LINE);
        if ((int)lptr >= (int)cptr)
            fprintf(outfile, "%s%s\t%s, temp%s", first, optr, pptr, NEW_LINE);
        else
            fprintf(outfile, "%s%s\ttemp,%s%s", first, optr, cptr, NEW_LINE);
    }

    return 1;
}

FILE *OpenFile(char *fname, char *mode)
{
    FILE *iofile;

    iofile = fopen(fname, mode);
    if (iofile) return iofile;

    printf("Could not open %s\n", fname);
    exit(1);
    return 0;
}

void RemoveCRLF(char *str)
{
    int len = strlen(str);
    str += len - 1;
    while (len-- && (*str == 10 || *str == 13)) *str-- = 0;
}

void ProcessComm(void)
{
    int len;
    char *ptr = FindChar(buffer1, ',');
    if (*ptr) *ptr++ = 0;
    fprintf(outfile, "%s%s", &buffer1[7], NEW_LINE);
    sscanf(ptr, "%d", &len);
    if (len <= 4)
        fprintf(outfile, "\tlong\t0%s", NEW_LINE);
    else
        fprintf(outfile, "\tlong\t0[%d]%s", (len+3)/4, NEW_LINE);
}

void ProcessZero(void)
{
    int len;
    sscanf(&buffer1[6], "%d", &len);
    if (len <= 4)
        fprintf(outfile, "\tlong\t0%s", NEW_LINE);
    else
        fprintf(outfile, "\tlong\t0[%d]%s", (len+3)/4, NEW_LINE);
}

void CheckLocalName(void)
{
    char *ptr = buffer1;

    while (*ptr)
    {
        ptr = FindChar(ptr, '.');
        if (!*ptr) break;
        if (ptr != buffer1 && isdigit((int)ptr[1]) &&
            (ptr[-1] == '_' || isdigit((int)ptr[-1]) || isalpha((int)ptr[-1]))) *ptr = '_';
        ptr++;
    }
}

int main(int argc, char **argv)
{
    char fname[100];
    char *extptr;
    FILE *prefile;

    if (argc != 2)
    {
        printf("usage: s2p filename\n");
        exit(1);
    }

    strcpy(fname, argv[1]);
    extptr = FindChar(fname, '.');
    if (!*extptr) strcpy(extptr, ".s");
    infile = OpenFile(fname, "r");
    strcpy(extptr, ".spin2");
    outfile = OpenFile(fname, "w");

    prefile = OpenFile("prefix.spin2", "r");
    while (fgets(buffer1, 1000, prefile))
    {
        RemoveCRLF(buffer1);
        fprintf(outfile, "%s%s", buffer1, NEW_LINE);
    }
    fclose(prefile);

    while (fgets(buffer1, 1000, infile))
    {
        RemoveCRLF(buffer1);
        if (!strncmp(buffer1, "\t.data", 6)) continue;
        if (!strncmp(buffer1, "\t.local", 7)) continue;
        if (!strncmp(buffer1, "\t.text", 6)) continue;
        if (!strncmp(buffer1, "\t.global", 8)) continue;
        if (!strncmp(buffer1, "\t.section", 9)) continue;
        if (!strncmp(buffer1, "\t.ascii \"", 8))
        {
            ProcessAscii();
            fprintf(outfile, "%s%s", buffer1, NEW_LINE);
            continue;
        }
        CheckLocalName();
        if (!strncmp(buffer1, "\t.comm\t", 7))
        {
            ProcessComm();
            continue;
        }
        if (!strncmp(buffer1, "\t.zero\t", 7))
        {
            ProcessZero();
            continue;
        }
        ReplaceString(".L", "_L");
        ReplaceString(".balign\t4", "alignl");
        ReplaceString(".balign\t2", "alignw");
        ReplaceString("0x", "$");
        ReplaceString("jmpret", "calld");
        ReplaceString("__MASK_", "##$");
        if (CheckMova()) continue;
        if (CheckWaitcnt()) continue;
        if (!strncmp(buffer1, "\tlong\t_L", 8))
        {
            fprintf(outfile, "\tlong\t%s%s", &buffer1[6], NEW_LINE);
            continue;
        }
        if (!strcmp(buffer1, "\tlong\t__clkfreq"))
        {
            fprintf(outfile, "\tlong\t%s%s", &buffer1[6], NEW_LINE);
            continue;
        }
	if (CheckSourceDest()) continue;
        if (CheckNR()) continue;
	if (CheckCNT()) continue;
        fprintf(outfile, "%s%s", buffer1, NEW_LINE);
    }
    fclose(infile);
    fclose(outfile);
    return 0;
}
